/***************************************************************************
                          ispellmanagerdlgimpl.cpp  -  description
                             -------------------
    begin                : Tue Oct 9 2001
    copyright            : (C) 2001 by Andriy Rysin
    email                : arysin@yahoo.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ispelldict.h"
#include <qwidget.h> 
#include <qfiledialog.h> 
#include <qlistbox.h>
#include <qbuttongroup.h>
#include <qtabwidget.h>
#include <qlabel.h>
#include <qcheckbox.h>
#include <qradiobutton.h>
#include <qpushbutton.h>
#include <qlineedit.h>

#include <fstream>
#include <iostream.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
	    
IspellDictDlgImpl::IspellDictDlgImpl(const char* fname_, QWidget *parent, const char *name ) : 
    IspellDictDlg(parent,name, true) 
{
    outputChanged = false;
    batchChanges = false;

    openFile(fname_);
}

IspellDictDlgImpl::~IspellDictDlgImpl()
{
}

void IspellDictDlgImpl::closeEvent ( QCloseEvent * e ) 
{
    // check for outputChanged

//    close(ispell_pipe);
    e->accept();
}

QString 
IspellDictDlgImpl::pipeIt(const char* args[], const QString& input) 
{
    if( pipe(topipe) == -1 || pipe(frompipe) == -1 ) {
	cerr << "pipes err" << endl;
    }

    char str[2096];

    pid_t subtask = fork();
    if( subtask == -1 ) {
	cerr << "Fork error!" << endl;
	exit(1);
    }

    if( subtask == 0 ) {
// child
	::close( frompipe[0]);
	::close( topipe[1]);
	dup2( topipe[0], 0); 
	dup2( frompipe[1], 1); 
	fcntl(0, F_GETFD, ~FD_CLOEXEC);
	fcntl(1, F_GETFD, ~FD_CLOEXEC);

	execvp(args[0], (char* const*)args);

	cerr << "exec error" << endl;
//	 signal(getppid(), SIGQUIT);
	exit(1);
    }

// parent
    ::close(frompipe[1]);
    ::close(topipe[0]);
    
    write(topipe[1], input.local8Bit(), input.length());
    write(topipe[1], "\n", 1);
    ::close(topipe[1]);
cerr << "waiting" << endl;
    int status;
	if( wait(&status) != subtask ) {
	    cerr << "Bad wait" << endl;
	    exit(1);
	}

    str[ read( frompipe[0], str, 2096) ] = '\0';
    
    ::close(frompipe[0]);
    
    return QString::fromLocal8Bit(str).stripWhiteSpace();
}

void
IspellDictDlgImpl::openFile(const QString& fname_)
{

	if( ! fname_.isEmpty() )
	    inputFile = fname_;
	else
	    inputFile = QString(getenv("HOME")) + "/.ispell_ukrainian";

	ifstream infile(inputFile);

//  infile.exceptions(ios::bad | ios::fail);

	if( !infile.good() ) {
	    cerr << "Can not open file " << inputFile << ", or it's empty!" << endl;
	    close();
	}

	lstInput->clear();

	char str[256];
	while( infile.good() )        {
	    infile.getline(str, 256);
	    if( *str != '\0' )
		lstInput->insertItem( QString::fromLocal8Bit(str).stripWhiteSpace() );
	}

        if( lstInput->count() > 0 ) {
    	    lstInput->sort();	
	    lstInput->setCurrentItem(0);
//	    wordSelected(lstInput->currentText());
	}
}

void
IspellDictDlgImpl::wordSelected() 
{
cerr << "word '" << lstInput->currentText().local8Bit() << "' selected" << endl;
    checkIspellVariants();
}

void
IspellDictDlgImpl::checkIspellVariants() 
{
cerr << "check ispell" << endl;
    lstSuggestions->clear();

    const char* args[] = {"ispell", "-d", "ukrainian", "-c", NULL};

    QString str = pipeIt(args, lstInput->currentText());

    int pr_pos = 0;
    int pos = 0;
    do {
	int len;
	pos = str.find(' ', pr_pos);
	if( pos == -1 )
	    len = -1;
	else
	    len = pos - pr_pos;
	    
    	lstSuggestions->insertItem( str.mid(pr_pos, len) );
	
	int ind = lstSuggestions->count() - 1;
	QString s = lstSuggestions->text(ind);
	if( s.find("/") != -1 )
	    s = s.right(4).left(2);
	else
	    s = s.right(2);
	if( s.local8Bit() == "" )
	    lstSuggestions->item(ind)->setSelectable(false);
	
	pr_pos = pos + 1;
    } while( pos != -1 );
    
    guessSuggGroup();

}

void 
IspellDictDlgImpl::parseSuggestion(const QString& str, QString& base_, QString& tag)
{
    int slash_pos = str.find("/");
    base_ = str.left(slash_pos);
    if( slash_pos != -1 )
	tag = str.right(str.length() - slash_pos - 1);	// actually "ispell -c" offers us only 1 flag
    else
	tag = "";
}


void 
IspellDictDlgImpl::guessSuggGroup()
{
cerr << "guess sugg group" << endl;
    
// coosing best offer from Ispell -c

    if( lstSuggestions->count() > 0 ) {
	lstSuggestions->setCurrentItem(0);
//	suggSelected();
    }
}


struct {
    const char* text;
    char flag;
    // boolean suffix;
    } guesses[] = 
{
    {"", ''},
    {"", ''},

    {"", 'a'},
    {"", 'a'},
    {"", 'a'},

    {"", 'e'},
    {"", 'e'},

    {"", 'i'},
    {"", 'i'},
    {"", 'i'},
    {"", 'i'},
    {"", 'i'},
    {"", 'i'},
    {"", 'i'},

    {"", 'l'},

    {"", 'V'},
//    {"", 'V'},

//    {"", 'i'},
    {"", 'a'},
    {"", 'a'},
    {"", 'e'},
    {"", 'e'},
    {"", 'e'},
    {"", 'e'},
    {"", 'W'},

    {"", 'A'},
    {0, 0}
};


void
IspellDictDlgImpl::suggSelected()
{
    parseSuggestion(lstSuggestions->currentText(), base, groupTag);
    char flag = '\0';

cerr << " sugg '" << base.local8Bit() << "' with flag '" << groupTag.latin1() << "' selected" << endl;

    if( groupTag.isEmpty() ) {	// searching for best infinitive form


	for( int ii=0; guesses[ii].text != 0; ii++) {
	    if( base.right(strlen(guesses[ii].text)) 
		    == QString::fromLocal8Bit(guesses[ii].text) ) {
		flag = guesses[ii].flag;
		break;
	    }
	}
//	else	// could not guess :(	
cerr << " flag's empty, best one seems to be " << flag << endl;

    }
    else
	flag = groupTag.at(0).latin1();
    
    batchChanges = true;

    if( strchr("abcdefghijklmnp", flag) )
	WordGroupTag->setCurrentPage(TAB_NOUN);
    else
    if( strchr("ABIJKLMNGHCDEF", flag) )
	WordGroupTag->setCurrentPage(TAB_VERB);
    else
    if( strchr("VWY", flag) )
	WordGroupTag->setCurrentPage(TAB_OTHER);

    switch ( flag )
    {
	case 'a':
	    chkSingle->setChecked(true);
	    rdGroup_a->setChecked(true);
	break;
	case 'c':
	    chkSingle->setChecked(true);
	    rdGroup_a->setChecked(true);
	    chkGen_a->setChecked(true);
	break;
	case 'b':
	    chkPlural->setChecked(true);
	    rdGroup_a->setChecked(true);
	    chkSwitch->setChecked(false);
	break;
	case 'o':
	    chkPlural->setChecked(true);
	    rdGroup_a->setChecked(true);
	    chkSwitch->setChecked(true);
	break;
	case 'd':
	    chkSingle->setChecked(true);
	    rdGroup_a->setChecked(true);
	    chkVocative->setChecked(true);
	break;
	
	case 'e':
	    chkSingle->setChecked(true);
	    rdGroup_e->setChecked(true);
	break;
	case 'f':
	    chkPlural->setChecked(true);
	    rdGroup_e->setChecked(true);
	break;
	case 'g':
	    chkSingle->setChecked(true);
	    rdGroup_e->setChecked(true);
	    chkSwitch->setChecked(true);
	break;
	case 'h':
	    chkSingle->setChecked(true);
	    rdGroup_e->setChecked(true);
	    chkVocative->setChecked(true);
	break;
	
	case 'i':
	    chkSingle->setChecked(true);
	    rdGroup_i->setChecked(true);
	break;
	case 'j':
	    chkPlural->setChecked(true);
	    rdGroup_i->setChecked(true);
	break;
	case 'k':
	    chkSingle->setChecked(true);
	    rdGroup_i->setChecked(true);
	    chkVocative->setChecked(true);
	break;
	
	case 'l':
	    chkSingle->setChecked(true);
	    rdGroup_l->setChecked(true);
	break;
	case 'm':
	    chkPlural->setChecked(true);
	    rdGroup_l->setChecked(true);
	break;
	case 'n':
	    chkSingle->setChecked(true);
	    rdGroup_l->setChecked(true);
	    chkVocative->setChecked(true);
	break;

	case 'p':
	    chkSingle->setChecked(true);
	    //rdGroup_l->setChecked(true); // צ 
	    chkPatronim->setChecked(true);
	break;

	case 'A':
	    chkNormal->setChecked(true);
	    rdGroupH->setChecked(true);
	break;
	case 'B':
	    chkReverse->setChecked(true);
	    rdGroupH->setChecked(true);
	break;
	case 'C':
	    chkNormal->setChecked(true);
	    rdImperSoft->setChecked(true);
	break;
	case 'D':
	    chkReverse->setChecked(true);
	    rdImperSoft->setChecked(true);
	break;
	case 'E':
	    chkNormal->setChecked(true);
	    rdImperHard->setChecked(true);
	break;
	case 'F':
	    chkReverse->setChecked(true);
	    rdImperHard->setChecked(true);
	break;
	case 'I':
	    chkNormal->setChecked(true);
	    rdGroupI->setChecked(true);
	break;
	case 'J':
	    chkReverse->setChecked(true);
	    rdGroupI->setChecked(true);
	break;
	case 'K':
	    chkNormal->setChecked(true);
	    rdGroupP->setChecked(true);
	break;
	case 'L':
	    chkReverse->setChecked(true);
	    rdGroupP->setChecked(true);
	break;
	case 'M':
	    chkNormal->setChecked(true);
	    rdGroupU->setChecked(true);
	break;
	case 'N':
	    chkReverse->setChecked(true);
	    rdGroupU->setChecked(true);
	break;

	case 'G':
	    chkNormal->setChecked(true);
	    chkComplexFuture->setChecked(true);
	break;
	case 'H':
	    chkReverse->setChecked(true);
	    chkComplexFuture->setChecked(true);
	break;

	case 'W':
	    chkAdverb->setChecked(true);
	    chkAdjective->setChecked(true);
	    chkCompar->setChecked(false);
	break;
	case 'Y':
	    chkCompar->setChecked(true);
	    chkAdjective->setChecked(true);
	    chkAdverb->setChecked(false);
	break;
	case 'V':
	    chkAdjective->setChecked(true);
	break;

	case 'Z':
	    chkNegative->setChecked(true);
	
    }
    
    batchChanges = false;

    updateGroupTag();
    updateRawText();
}

void
IspellDictDlgImpl::updateRawText() 
{
    if( batchChanges )
	return;

    QString full_word = base;
    if( ! groupTag.isEmpty() ) {
	full_word += "/";
	full_word += groupTag;
    }
    txtResult->setText( full_word );
    
    btnDecline->setEnabled(false);
    getDeclinations();
}

/*
    gets all word forms genrated with given flags
*/

void
IspellDictDlgImpl::getDeclinations()
{
cerr << "get declinations" << endl;

    lstDeclinations->clear();

    QString raw_text = txtResult->text();
    
    if( raw_text.find("/") == -1 )
	return;

    const char* args[] = {"ispell", "-d", "ukrainian", "-e", NULL};

    QString str = pipeIt(args, raw_text);

    int pr_pos = 0;
    int pos = 0;
    do {
	int len;
	pos = str.find(' ', pr_pos);
	if( pos == -1 )
	    len = -1;
	else
	    len = pos - pr_pos;
	    
    	lstDeclinations->insertItem( str.mid(pr_pos, len) );
	pr_pos = pos + 1;
    } while( pos != -1 );
}

void
IspellDictDlgImpl::groupChanged( int ) 
{
/*    if( chkNotDeclined->isChecked() ) {
	chkAdjective->setChecked(false);
	chkAdverb->setChecked(false);
	chkNegative->setChecked(false);
    }
*/
    if( batchChanges )
	return;
	
    updateGroupTag();
    updateRawText();
}

/*
    setting initial buttons states for selected tab
*/

void
IspellDictDlgImpl::tabChanged( const QString& tabname ) 
{
    if( batchChanges )
	return;

    int tab = WordGroupTag->currentPageIndex();

    batchChanges = true;
    
    chkNegative->setChecked(false);
    
    switch ( tab ) {
    case TAB_NOUN:
	chkSingle->setChecked(false);
	chkPlural->setChecked(false);
	chkSwitch->setChecked(false);
	chkGen_a->setChecked(false);
	chkVocative->setChecked(false);
	chkPatronim->setChecked(false);
	rdGroup_a->setChecked(true);
    break;
    case TAB_VERB:
	chkReverse->setChecked(false);
	chkComplexFuture->setChecked(false);
	rdGroupH->setChecked(true);
    break;
    case TAB_OTHER:
	chkAdverb->setChecked(false);
	chkCompar->setChecked(false);
	chkAdjective->setChecked(true); 
    break;
    }

    batchChanges = false;
    
    groupTag = "";
    updateRawText();
}

/*
    updates groupTag in the text field from current positions of tab and radio/check buttons
*/

void
IspellDictDlgImpl::updateGroupTag( ) 
{
// direct, reverse
    static const char verb_pairs[4][2] = { {'A', 'B'}, {'I', 'J'}, {'K', 'L'}, {'M', 'N'} };
// singular, plural, vocative
    static const char noun_pairs[4][3] = { {'a', 'b', 'd'}, {'e', 'f', 'h'}, {'i', 'j', 'k'}, {'l', 'm', 'n'} };
    
    int tab = WordGroupTag->currentPageIndex();
    
    groupTag = "";
    
    switch ( tab ) {
    
    case TAB_NOUN:
    {
	int noun_type = grpNounTypes->id(grpNounTypes->selected());
    
	if( chkSingle->isChecked() ) {
		groupTag += noun_pairs[ noun_type ][0];
		if( chkGen_a->isChecked() )
		    if( noun_type == 0 ) groupTag += "c"; // ?? 'c'
		    else
			if( noun_type == 1 ) groupTag += "g";
	}
	if( chkPlural->isChecked() ) {
		if( rdGroup_a->isChecked() && chkSwitch->isChecked() )  // plural "b" or "o"
		    groupTag += "o";
		else
		    groupTag += noun_pairs[ noun_type ][1];
	}
	if( chkVocative->isChecked() )
		groupTag += noun_pairs[ noun_type ][2];
	if( chkPatronim->isChecked() )
		groupTag += "p";
    }
    break;
    
    case TAB_VERB:
	{
	int verb_type = grpVerbTypes->id(grpVerbTypes->selected());
	if( chkNormal->isChecked() ) {
	    groupTag += verb_pairs[ verb_type ][0];
	    if( chkComplexFuture->isChecked() )
		groupTag += 'G';
	    if( verb_type == 0 ) {
		if( rdImperHard->isChecked() )
		    groupTag += 'C';
		else
		if( rdImperSoft->isChecked() )
		    groupTag += 'E';
	    }
	}
    
	if( chkReverse->isChecked() ) {
		groupTag += verb_pairs[ verb_type ][1];
	    if( chkComplexFuture->isChecked() )
		groupTag += 'H';
	    if( verb_type == 0 ) {
		if( rdImperHard->isChecked() )
		    groupTag += 'D';
		else
		if( rdImperSoft->isChecked() )
		    groupTag += 'F';
	    }
	}
	}
    break;
    
    case TAB_OTHER:
	if( chkAdjective->isChecked() ) {
	    groupTag += "V"; 
	    if( chkAdverb->isChecked() ) {
		groupTag += "W";
	    } else
	    if( chkCompar->isChecked() ) {
		groupTag += "Y";
	    }
	    
	}
    break;
    }
    
    if( chkNegative->isChecked() ) {
	groupTag += "Z";
    }
}

void
IspellDictDlgImpl::rawTextEdited() 
{
    btnDecline->setEnabled(true);
    QString str = txtResult->text();
    int pos = str.find("/");
    if( pos == -1 )
	base = str;
    else
	base = str.left(pos);
}

void
IspellDictDlgImpl::addWord() 
{
    lstOutput->insertItem( txtResult->text().stripWhiteSpace() );
    int curr_item = lstInput->currentItem();
    lstInput->removeItem( curr_item );
    if( chkSortedOutput->isChecked() == true )
        lstOutput->sort();
	
    if( lstInput->count() > curr_item )
	lstInput->setCurrentItem( curr_item );
    else
	if( lstInput->count() > 0 )
	    lstInput->setCurrentItem( lstInput->count() - 1 );
	
    outputChanged = true;
}

void 
IspellDictDlgImpl::removeWord() 
{
    lstInput->removeItem( lstInput->currentItem() );
}

void 
IspellDictDlgImpl::duplicateWord() 
{
    lstInput->insertItem( lstInput->currentText(), lstInput->currentItem() );
}

void 
IspellDictDlgImpl::getWordBack() 
{
    if( lstOutput->currentItem() != -1 ) {
	QString str =  lstOutput->currentText();
	str.truncate( str.findRev("/") );
	lstInput->insertItem( str );
	lstInput->sort();
	lstInput->setCurrentItem( lstInput->findItem( str ) );
	lstOutput->removeItem( lstOutput->currentItem() );
    }
}

void
IspellDictDlgImpl::save()
{
    // check for outputChanged

    QString fileName = QFileDialog::getSaveFileName( "newfile.txt", "Textfiles (*.txt)", this );
    if ( !fileName.isNull() ) {                 // got a file name
       
	// writing output
	ofstream outfile( fileName );
        for( uint ii=0; ii<lstOutput->count(); ii++) {
	    outfile << (lstOutput->text( ii )).local8Bit() << endl;
        }
	if( chkSaveInput->isChecked() ) {
	    // rewriting input
	    ofstream infile( inputFile );
    	    for( uint ii=0; ii<lstInput->count(); ii++) {
	        infile << (lstInput->text( ii )).local8Bit() << endl;
    	    }
	}
	
    }
    else { // error msg
    }
}		   

void
IspellDictDlgImpl::open()
{
 QString fname( QFileDialog::getOpenFileName( QString::null, "All files (*)", this ) );
 if ( fname.isEmpty() )
       return;

    openFile( fname );       
}
